## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

##"""
## ___      ___ ___ ___   _     _   _ 
##| _ \_  _/ __| __/ __| /_\   (_) (_)
##|  _/ || \__ \ _|\__ \/ _ \   _   _ 
##|_|  \_, |___/___|___/_/ \_\ (_) (_)
##     |__/                           
##                        __
##   ________  ____ _____/ /
##  / ___/ _ \/ __ `/ __  / 
## / /  /  __/ /_/ / /_/ /  
##/_/   \___/\__,_/\__,_/   


##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+

##"""

# import libraries
import numpy as np
cimport numpy as np
cimport cython

# =========================================================
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.nonecheck(False)
cpdef np.ndarray txtread(str infile):
   '''
   Custom fast (up to 3.5x faster than numpy's genfromtxt) txt file to numpy array
   accepts comma, tab or space delimited files of 3 columns: x, y, and amplitude


   Syntax
   ----------
   pts = pysesa_read.txtread(infile)

   Parameters
   ------------
   infile : str
   	3-column ASCII file containing Nx3 point cloud

   Returns
   ----------
   data: ndarray
   	Nx3 point cloud, 32 bit precision

   '''
   f = open(infile, 'rb'); data = f.read(); f.close()
   data = data.splitlines()
   
   cdef np.ndarray[np.float32_t, ndim=2] out = np.empty((len(data),3),dtype=np.float32)
   
   if len(data[0].split(','))>1: # the file is comma delimited
      out = np.array([x.split(',',2)[0:3] for x in data], dtype='float32')
      return out
   else:
      try: # space delimited
         out = np.array([x.split(' ',2)[0:3] for x in data], dtype='float32')
         return out
      except: #tab delimited
         out = np.array([x.split('\t',2)[0:3] for x in data], dtype='float32')
         return out
         



# encoding: utf-8
## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
PySESA - a Python framework for Spatially Explicit Spectral Analysis

PySESA is an open-source project dedicated to provide a generic Python framework 
for spatially explicit statistical analyses of point clouds and other geospatial data, 
in the spatial and frequency domains, for use in the geosciences

The program is detailed in:
   Buscombe, D. (2016) "Computational considerations for spatially explicit spectral analysis of point clouds and geospatial data", 86, 92-108, 10.1016/j.cageo.2015.10.004.

:Author:  
    Daniel Buscombe
    Grand Canyon Monitoring and Research Center
    United States Geological Survey
    Flagstaff, AZ 86001
    dbuscombe@usgs.gov

For more information visit http://dbuscombe-usgs.github.io/pysesa/

:install:
    python setup.py install
    sudo python setup.py install
    
:test:
    python -c "import pysesa; pysesa.test()"

:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
    
    This software is in the public domain because it contains materials that
    originally came from the United States Geological Survey, an agency of the
    United States Department of Interior. For more information, 
    see the official USGS copyright policy at
    http://www.usgs.gov/visual-id/credit_usgs.html#copyright
    Any use of trade, product, or firm names is for descriptive purposes only 
    and does not imply endorsement by the U.S. government.
    
"""

__version__ = '0.0.28'

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from pysesa._pysesa import process
from pysesa.test import *

import pysesa.read
import pysesa.write
from pysesa.partition import partition
from pysesa.sgolay import sgolay
from pysesa.spatial import spatial
from pysesa.spectral import spectral
from pysesa.lengthscale import lengthscale
from pysesa.detrend import detrend

from pysesa.plot import plot





## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitoring & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
PySESA - a Python framework for Spatially Explicit Spectral Analysis

PySESA is an open-source project dedicated to provide a generic Python framework 
for spatially explicit statistical analyses of point clouds and other geospatial data, 
in the spatial and frequency domains, for use in the geosciences

The program is detailed in:
   Buscombe, D. (2016) "Computational considerations for spatially explicit spectral analysis of point clouds and geospatial data", 86, 92-108, 10.1016/j.cageo.2015.10.004.

:Author:  
    Daniel Buscombe
    Grand Canyon Monitoring and Research Center
    United States Geological Survey
    Flagstaff, AZ 86001
    dbuscombe@usgs.gov

For more information visit http://dbuscombe-usgs.github.io/pysesa/

:install:
    python setup.py install
    sudo python setup.py install
    
:test:
    python -c "import pysesa; pysesa.test()"

:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
    
    This software is in the public domain because it contains materials that
    originally came from the United States Geological Survey, an agency of the
    United States Department of Interior. For more information, 
    see the official USGS copyright policy at
    http://www.usgs.gov/visual-id/credit_usgs.html#copyright
    Any use of trade, product, or firm names is for descriptive purposes only 
    and does not imply endorsement by the U.S. government.
    
"""

# =========================================================
# ======================== libraries ======================
# =========================================================

from __future__ import division
import numpy as np
from joblib import Parallel, delayed, cpu_count
from time import clock, time
import os #, sys, getopt

# suppress divide and invalid warnings
np.seterr(divide='ignore')
np.seterr(invalid='ignore')
np.seterr(over='ignore')
np.seterr(under='ignore')

import pysesa

import warnings
warnings.filterwarnings("ignore")

# =========================================================
# ===================== subfunctions ======================
# =========================================================

# =========================================================
def ascol( arr ):
   '''
   reshapes row matrix to be a column matrix (N,1).
   '''
   if len( arr.shape ) == 1: arr = arr.reshape( ( arr.shape[0], 1 ) )
   return arr

#==================================================
def get_spec(pts, spectype, out, detrend, res, method, nbin, lentype, taper):
   '''
   call the spectral analysis routine for detrended window of point cloud
   Gets called by the parallel processing queue
   '''
   try:   
      pts = pysesa.detrend(pts, detrend, res, method).getdata()
      return pysesa.spatial(pts).getcentroid() + pysesa.spectral(pts, nbin, res, spectype, lentype, taper, method).getdata()
   except:
      return [np.ones(27)*np.nan]
      
#==================================================
def get_spat(pts, detrend, res, method):
   '''
   call the spatial analysis routine for detrended window of point cloud
   Gets called by the parallel processing queue
   '''
   try:
      pts = pysesa.detrend(pts, detrend, res, method).getdata()
      return pysesa.spatial(pts).getdata()
   except:
      return [np.ones(10)*np.nan]
      
#==================================================
def get_spec_spat(pts, spectype, out, detrend, res, method, nbin, lentype, taper):
   '''
   call the spectral and spatial analysis routine for detrended window of point cloud
   Gets called by the parallel processing queue
   '''
   try:
      pts = pysesa.detrend(pts, detrend, res, method).getdata()
      return pysesa.spatial(pts).getdata() + pysesa.spectral(pts, nbin, res, spectype, lentype, taper, method).getdata()
   except:
      return [np.ones(34)*np.nan]

# =========================================================
# ==================== begin program ======================
# =========================================================
def process(infile, out=1, detrend=4, proctype=1, mxpts=1024, res=0.05, nbin=20, lentype=1, minpts=64, taper=1, prc_overlap=0): # bp=0
   '''
   Calculate spectral and spatial statistics of a Nx3 point cloud

   Syntax
   ----------
   () = pysesa.process(infile, out, detrend, proctype, mxpts, res, nbin, lentype, minpts, taper, prc_overlap)

   Parameters
   -----------
   infile : str
   	ASCII file containing an Nx3 point cloud in 3 columns

   Other Parameters
   -----------------
   out : float, *optional* [default = 0.5]
   	output grid resolution
   detrend : int, *optional* [default = 4]
   	type of detrending.
        1 = remove mean
        2 = remove Ordinary least squares plane
        3 = remove Robust linear model plane
        4 = remove Orthogonal Distance Regression plane
        5 = remove Savitsky-Golay digital filter, order 1
   proctype : int, *optional* [default = 1, no spectral smoothing]
   	proctype type:
        1 = spectral only, no spectral smoothing
        2 = spectral only, spectrum smoothed with Gaussian
        3 = spatial only
        4 = spatial + spectrum, no spectral smoothing
        5 = spatial + spectrum smoothed with Gaussian
   mxpts : float, *optional* [default = 1024]
   	maximum number of points allowed in a window
   res : float, *optional* [default = 0.05]
        spatial grid resolution to create a grid
   nbin : int, *optional* [default = 20]
        number of bins for power spectral binning
   lentype : int, *optional* [default = 1, l<0.5]
   	lengthscale type:
        1 = l<0.5
        2 = l<1/e
        3 = l<0
   minpts : float, *optional* [default = 16]
   	minimum number of points allowed in a window
   taper : int, *optional* [default = Hanning]
   	flag for taper type:
        1 = Hanning (Hann)
        2 = Hamming
        3 = Blackman
        4 = Bartlett
   prc_overlap : float, *optional"  [default = 0]
        percentage overlap between windows


   Returns [proctype = 1 or proctype = 2]
   ---------------------------------------
   data: list
   	x = centroid in horizontal coordinate
        y = centroid in laterial coordinate
        z = centroid in vertical coordinate
   	slope = slope of regression line through log-log 1D power spectral density
        intercept = intercept of regression line through log-log 1D power spectral density
        r_value = correlation of regression through log-log 1D power spectral density
        p_value = probability that slope of regression through log-log 1D power spectral density is not zero
        std_err = standard error of regression through log-log 1D power spectral density
        d = fractal dimension
        l = integral lengthscale
        wmax = peak wavelength
        wmean = mean wavelength
        rms1 = RMS amplitude from power spectral density
        rms2 = RMS amplitude from bin averaged power spectral density
        Z = zero-crossings per unit length
        E = extreme per unit length
        sigma = RMS amplitude
        T0_1 = average spatial period (m_0/m_1)
        T0_2 = average spatial period (m_0/m_2)^0.5
        sw1 = spectral width 
        sw2 = spectral width (normalised radius of gyration)
        m0 = zeroth moment of spectrum
        m1 = first moment of spectrum
        m2 = second moment of spectrum
        m3 = third moment of spectrum
        m4 = fourth moment of spectrum
        phi = effective slope (degrees)

   Returns [proctype = 3]
   -------------------------
   data: list
   	x = centroid in horizontal coordinate
        y = centroid in laterial coordinate
        z = centroid in vertical coordinate
        z_mean = centroid in amplitude
        z_max = max amplitude
        z_min = min amplitude
        z_range = range in amplitude
        sigma = standard deviation of amplitudes
        skewness = skewness of amplitudes
        kurtosis = skewness of amplitudes
        n = number of 3D coordinates

   Returns [proctype = 4 or proctype = 5]
   -----------------------------------------
   data: list
   	x = centroid in horizontal coordinate
        y = centroid in laterial coordinate
        z_mean = centroid in amplitude
        z_max = max amplitude
        z_min = min amplitude
        z_range = range in amplitude
        sigma = standard deviation of amplitudes
        skewness = skewness of amplitudes
        kurtosis = skewness of amplitudes
        n = number of 3D coordinates
   	slope = slope of regression line through log-log 1D power spectral density
        intercept = intercept of regression line through log-log 1D power spectral density
        r_value = correlation of regression through log-log 1D power spectral density
        p_value = probability that slope of regression through log-log 1D power spectral density is not zero
        std_err = standard error of regression through log-log 1D power spectral density
        d = fractal dimension
        l = integral lengthscale
        wmax = peak wavelength
        wmean = mean wavelength
        rms1 = RMS amplitude from power spectral density
        rms2 = RMS amplitude from bin averaged power spectral density
        Z = zero-crossings per unit length
        E = extreme per unit length
        sigma = RMS amplitude
        T0_1 = average spatial period (m_0/m_1)
        T0_2 = average spatial period (m_0/m_2)^0.5
        sw1 = spectral width 
        sw2 = spectral width (normalised radius of gyration)
        m0 = zeroth moment of spectrum
        m1 = first moment of spectrum
        m2 = second moment of spectrum
        m3 = third moment of spectrum
        m4 = fourth moment of spectrum
        phi = effective slope (degrees)

   '''

   print """
       ..  _ __  _   _   ___  ___  ___  __ _ 
       .. | '_ \| | | | / __|/ _ \/ __|/ _` |
       .. | |_) | |_| | \__ \  __/\__ \ (_| |
       .. | .__/ \__, | |___/\___||___/\__,_|
       .. |_|    |___/ 
       .. 
       .. +-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
       .. |b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
       .. +-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
       .. +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       .. |d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
       .. +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       .. +-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
       .. |U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
       .. +-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
   """

   if infile:
      print 'Input file is %s' % (infile)

   if prc_overlap:
      prc_overlap = np.asarray(prc_overlap,int)
      print 'Percent overlap is %s' % (str(prc_overlap))
   elif prc_overlap==0:
      prc_overlap = np.asarray(prc_overlap,int)
      print 'Percent overlap is %s' % (str(prc_overlap))

   if out:
      out = np.asarray(out,float)
      print 'Output grid size is %s' % (str(out))

   if detrend:
      detrend = np.asarray(detrend,int)

      if detrend==1:
         print 'Detrend type: remove mean'
      if detrend==2:
         print 'Detrend type: Ordinary least squares plane'
      if detrend==3:
         print 'Detrend type: Robust linear model plane'
      if detrend==4:
         print 'Detrend type: Orthogonal Distance Regression plane'
      if detrend==5:
         print 'Detrend type: Savitsky-Golay digital filter, order 1'

   if proctype:
      proctype = np.asarray(proctype,int)
      if proctype==0:
         proctype = 1

      if proctype==1:
         print 'spectral parameters (no smoothing)'
      elif proctype==2:
         print 'spectral parameters (with smoothing)'
      elif proctype==3:
         print 'spatial parameters'
      elif proctype==4:
         print 'spatial parameters + spectral parameters (no smoothing)'
      elif proctype==5:
         print 'spatial parameters + spectral parameters (smoothing)'

   if res:
      res = np.asarray(res,float)
      print 'Res. is %s' % (str(res))

   if mxpts:
      mxpts = np.asarray(mxpts,int)
      print 'Max points per window is %s' % (str(mxpts))

   if minpts:
      minpts = np.asarray(minpts,int)
      print 'Min points per window is %s' % (str(minpts))

   if nbin:
      nbin = np.asarray(nbin,float)
      print 'Number of bins is %s' % (str(nbin))

   if lentype:
      lentype = int(lentype)
      if lentype==1:
         print "lengthscale type: l<0.5"
      elif lentype==2:
         print "lengthscale type: l<1/e"
      elif lentype==3:
         print "lengthscale type: l<0"

   if taper:
      taper = np.asarray(taper,int)
      if taper==1:
         print 'Hanning taper'
      elif taper==2:
         print 'Hamming taper'
      elif taper==3:
         print 'Blackman taper'
      else:
         print 'Bartlett taper'

   # start timer
   if os.name=='posix': # true if linux/mac or cygwin on windows
      start1 = time()
   else: # windows
      start1 = clock()

   method = 'nearest'

   #==============================================================================
   print "(1) Reading data from file ..."
   # read in ascii 3-column file containing point cloud data
   toproc = pysesa.read.txtread(infile)

   ## number of points, undecimated
   orig_pts = len(toproc)

   #==============================================================================
   print "(2) Partitioning data into windows ... " 
   # get indices to windows
   nr_pts = pysesa.partition(toproc, out, mxpts, minpts, prc_overlap).getdata() #res, bp

   #==============================================================================
   print "(3) Processing in parallel using %s processors ... " % (str(cpu_count()))

   #==============================================================================
   if (proctype==1) or (proctype==2):

      #spectral, no smooth
      if proctype==1:
         try: #parallel processing with all available cores
            w = Parallel(n_jobs=cpu_count(), verbose=0)(delayed(get_spec)(toproc[nr_pts[k],:3], 1, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts))) 
         except: #fall back to serial
            w = Parallel(n_jobs=1, verbose=0)(delayed(get_spec)(toproc[nr_pts[k],:3], 1, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts)))

      #spectral, with smooth
      if proctype==2:
         try: #parallel processing with all available cores
            w = Parallel(n_jobs=cpu_count(), verbose=0)(delayed(get_spec)(toproc[nr_pts[k],:3], 2, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts))) 
         except: #fall back to serial
            w = Parallel(n_jobs=1, verbose=0)(delayed(get_spec)(toproc[nr_pts[k],:3], 2, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts)))

      try:
         x, y, z, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi = zip(*w)
      except:
         w2 = []
         for k in xrange(len(w)):
            if len(w[k])==27:
               w2.append(w[k])
         x, y, z, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi = zip(*w2)
         del w2   

      del w

      # combine into single matrix for writing to file
      towrite = np.hstack(( ascol(np.asarray(x)),ascol(np.asarray(y)),ascol(np.asarray(z)),ascol(np.asarray(slope)),ascol(np.asarray(intercept)),ascol(np.asarray(r_value)),ascol(np.asarray(p_value)),ascol(np.asarray(std_err)),ascol(np.asarray(d)),ascol(np.asarray(l)),ascol(np.asarray(wmax)),ascol(np.asarray(wmean)),ascol(np.asarray(rms1)),ascol(np.asarray(rms2)),ascol(np.asarray(Z)),ascol(np.asarray(E)),ascol(np.asarray(sigma)),ascol(np.asarray(T0_1)),ascol(np.asarray(T0_2)),ascol(np.asarray(sw1)),ascol(np.asarray(sw2)),ascol(np.asarray(m0)),ascol(np.asarray(m1)),ascol(np.asarray(m2)),ascol(np.asarray(m3)),ascol(np.asarray(m4)),ascol(np.asarray(phi)) ))

      del x, y, z, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi

      # remove rows with any NaNs
      towrite = towrite[np.where(np.logical_not(np.any(np.isnan(towrite),axis=1)))[0],:]

      # make a header string for the output file
      header = 'x, y, z, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi'

   #==============================================================================
   elif proctype==3: #spatial only

      try: #parallel processing with all available cores
         w = Parallel(n_jobs=cpu_count(), verbose=0)(delayed(get_spat)(toproc[nr_pts[k],:3], detrend, res, method) for k in xrange(len(nr_pts))) 
      except: #fall back to serial
         w = Parallel(n_jobs=1, verbose=0)(delayed(get_spat)(toproc[nr_pts[k],:3], detrend, res, method) for k in xrange(len(nr_pts)))

      try:
         x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n = zip(*w)
      except:
         w2 = []
         for k in xrange(len(w)):
            if len(w[k])==10:
               w2.append(w[k])
         x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n = zip(*w2)
         del w2   

      del w

      # combine into single matrix for writing to file
      towrite = np.hstack(( ascol(np.asarray(x)),ascol(np.asarray(y)),ascol(np.asarray(z_mean)),ascol(np.asarray(z_max)),ascol(np.asarray(z_min)),ascol(np.asarray(z_range)),ascol(np.asarray(sigma)),ascol(np.asarray(skewness)),ascol(np.asarray(kurtosis)), ascol(np.asarray(n)) ))

      del x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n 

      # remove rows with any NaNs
      towrite = towrite[np.where(np.logical_not(np.any(np.isnan(towrite),axis=1)))[0],:]

      # make a header string for the output file
      header = 'x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n'

   #==============================================================================
   elif (proctype==4) or (proctype==5): #spectral and spatial

      #spatial + spectral, no smooth
      if proctype==4:
         try: #parallel processing with all available cores
            w = Parallel(n_jobs=cpu_count(), verbose=0)(delayed(get_spec_spat)(toproc[nr_pts[k],:3], 1, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts))) 
         except: #fall back to serial
            w = Parallel(n_jobs=1, verbose=0)(delayed(get_spec_spat)(toproc[nr_pts[k],:3], 1, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts)))

      #spatial + spectral, with smooth
      if proctype==5:
         try: #parallel processing with all available cores
            w = Parallel(n_jobs=cpu_count(), verbose=0)(delayed(get_spec_spat)(toproc[nr_pts[k],:3], 2, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts))) 
         except: #fall back to serial
            w = Parallel(n_jobs=1, verbose=0)(delayed(get_spec_spat)(toproc[nr_pts[k],:3], 2, out, detrend, res, method, nbin, lentype, taper) for k in xrange(len(nr_pts)))

      try:
         x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi = zip(*w)
      except:
         w2 = []
         for k in xrange(len(w)):
            if len(w[k])==34:
               w2.append(w[k])
         x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi = zip(*w2)
         del w2   

      del w

      # combine into single matrix for writing to file
      towrite = np.hstack(( ascol(np.asarray(x)),ascol(np.asarray(y)), ascol(np.asarray(z_mean)),ascol(np.asarray(z_max)),ascol(np.asarray(z_min)),ascol(np.asarray(z_range)),ascol(np.asarray(sigma)),ascol(np.asarray(skewness)),ascol(np.asarray(kurtosis)), ascol(np.asarray(n)), ascol(np.asarray(slope)),ascol(np.asarray(intercept)),ascol(np.asarray(r_value)),ascol(np.asarray(p_value)),ascol(np.asarray(std_err)),ascol(np.asarray(d)),ascol(np.asarray(l)),ascol(np.asarray(wmax)),ascol(np.asarray(wmean)),ascol(np.asarray(rms1)),ascol(np.asarray(rms2)),ascol(np.asarray(Z)),ascol(np.asarray(E)),ascol(np.asarray(sigma)),ascol(np.asarray(T0_1)),ascol(np.asarray(T0_2)),ascol(np.asarray(sw1)),ascol(np.asarray(sw2)),ascol(np.asarray(m0)),ascol(np.asarray(m1)),ascol(np.asarray(m2)),ascol(np.asarray(m3)),ascol(np.asarray(m4)),ascol(np.asarray(phi)) ))

      del x, y, z_mean, z_max, z_min, z_range,  skewness, kurtosis, n #,sigma 
      del slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi

      # remove rows with any NaNs
      towrite = towrite[np.where(np.logical_not(np.any(np.isnan(towrite),axis=1)))[0],:]

      # make a header string for the output file
      header = 'x, y, z_mean, z_max, z_min, z_range, sigma, skewness, kurtosis, n, slope, intercept, r_value, p_value, std_err, d, l, wmax, wmean, rms1, rms2, Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4, phi'


   #==============================================================================
   print "(4) Writing data to file ..."

   # create a string for the output file
   outfile = infile+'_zstat_detrend'+str(detrend)+'_outres'+str(out)+'_proctype'+str(proctype)+'_mxpts'+str(mxpts)+'_minpts'+str(minpts)+'.xyz' 

   try:
      # write the data to the file
      pysesa.write.txtwrite(outfile, towrite, header)
   
   except:
      with open(outfile, 'wb') as f:
         np.savetxt(f, towrite[np.where(towrite[:,-1])[0],:], header = header, fmt=' '.join(['%8.6f,'] * np.shape(towrite)[1])[:-1]) 

   # stop the clock
   if os.name=='posix': # true if linux/mac
      elapsed = (time() - start1)
   else: # windows
      elapsed = (clock() - start1)

   print "Done! %s points decimated to %s points. Program ran for %s seconds" % (str(orig_pts), str(len(towrite)), str(elapsed))        




## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

##"""                       
## ___      ___ ___ ___   _     _   _ 
##| _ \_  _/ __| __/ __| /_\   (_) (_)
##|  _/ || \__ \ _|\__ \/ _ \   _   _ 
##|_|  \_, |___/___|___/_/ \_\ (_) (_)
##     |__/                           
##                         __             __
##   _________  ___  _____/ /__________ _/ /
##  / ___/ __ \/ _ \/ ___/ __/ ___/ __ `/ / 
## (__  ) /_/ /  __/ /__/ /_/ /  / /_/ / /  
##/____/ .___/\___/\___/\__/_/   \__,_/_/   
##    /_/                                   

##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##"""

# import libraries
from __future__ import division
import numpy as np
cimport numpy as np
cimport cython
from scipy.integrate import trapz
from scipy.stats import linregress
import statsmodels.api as smapi

#from libc.math cimport sqrt,abs

# import PySESA libraries
import pysesa.lengthscale

# suppress divide and invalid warnings
np.seterr(divide='ignore')
np.seterr(invalid='ignore')
np.seterr(over='ignore')
np.seterr(under='ignore')
import warnings
warnings.filterwarnings("ignore")

# =========================================================
cdef class spectral:
   '''
   Calculate spectral statistics of a Nx3 point cloud

   Syntax
   ----------
   data = pysesa.spectral.spec(points, nbin, res, proctype, lentype, taper, method).getdata()

   lengths = pysesa.spectral.spec(points, nbin, res, proctype, lentype, taper, method).getlengths()

   psdparams= pysesa.spectral.spec(points, nbin, res, proctype, lentype, taper, method).getstats()

   lengthscale = pysesa.spectral.spec(points, nbin, res, proctype, lentype, taper, method).getlengthscale()

   moments = pysesa.spectral.spec(points, nbin, res, proctype, lentype, taper, method).getmoments()

   Parameters
   ------------
   points : ndarray
   	Nx3 point cloud

   Other Parameters
   -------------------
   nbin : int, *optional* [default = 20]
        number of bins for power spectral binning
   res : float, *optional* [default = 0.05]
        spatial grid resolution to create a grid
   proctype : int, *optional* [default = 1, no spectral smoothing]
   	proctype type:
        1, no spectral smoothing

        2, spectrum smoothed with Gaussian

   lentype : int, *optional* [default = 1, l<0.5]
   	lengthscale type:
        1, l<0.5

        2, l<1/e

        3, l<0

   taper : int, *optional* [default = Hanning]
   	flag for taper type:
        1, Hanning (Hann)

        2, Hamming

        3, Blackman

        4, Bartlett

   method : str, *optional* [default = 'nearest']
   	gridding type

   Returns [requested through .getdata()]
   ----------------------------------------
   self.data: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension

        l = integral lengthscale

        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density

        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

        phi = effective slope (degrees)

   Returns [requested through .getpsdparams()]
   --------------------------------------------
   self.psdparams: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension

   Returns [requested through .getlengths()]
   -------------------------------------------
   self.lengths: list
        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density


   Returns [requested through .getlengthscale()]
   ----------------------------------------------
   self.lengthscale: float
        l = integral lengthscale

   Returns [requested through .getmoments()]
   ------------------------------------------
   self.moments: list
        Z = zero-crossings per unit length

        E = extreme per unit length
        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

        phi = effective slope (degrees)

   '''
   cdef object data, x_space, k_space, r, moments, lengthscale, lengths, psdparams

   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   # =========================================================
   def __init__(self, np.ndarray[np.float64_t, ndim=2] points, int nbin=20, double res=0.05, int proctype=1, int lentype=1, int taper=1, str method='nearest'): 
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      data = pysesa.spectral(points, nbin, res, proctype, lentype, taper, method).getdata()

      lengths = pysesa.spectral(points, nbin, res, proctype, lentype, taper, method).getlengths()

      psdparams= pysesa.spectral(points, nbin, res, proctype, lentype, taper, method).getstats()

      lengthscale = pysesa.spectral(points, nbin, res, proctype, lentype, taper, method).getlengthscale()

      moments = pysesa.spectral(points, nbin, res, proctype, lentype, taper, method).getmoments()

      Parameters
      ------------
      points : ndarray
   	Nx3 point cloud

      Other Parameters
      ------------------
      nbin : int, *optional* [default = 20]
        number of bins for power spectral binning
      res : float, *optional* [default = 0.05]
        spatial grid resolution to create a grid
      proctype : int, *optional* [default = 1, no spectral smoothing]
   	proctype type:
        1, no spectral smoothing

        2, spectrum smoothed with Gaussian

      lentype : int, *optional* [default = 1, l<0.5]
   	lengthscale type:
        1, l<0.5

        2, l<1/e

        3, l<0

      taper : int, *optional* [default = Hanning]
   	flag for taper type:
        1, Hanning (Hann)

        2, Hamming

        3, Blackman

        4, Bartlett

      method : str, *optional* [default = 'nearest']
   	gridding type

      Returns [requested through .getdata()]
      ---------------------------------------
      self.data: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension

        l = integral lengthscale

        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density

        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

        phi = effective slope (degrees)

      Returns [requested through .getpsdparams()]
      ---------------------------------------------
      self.psdparams: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension


      Returns [requested through .getlengths()]
      -------------------------------------------
      self.lengths: list
        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density


      Returns [requested through .getlengthscale()]
      -----------------------------------------------
      self.lengthscale: float
        l = integral lengthscale

      Returns [requested through .getmoments()]
      -------------------------------------------
      self.moments: list
        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

      '''

      cdef float pi = 3.14159265

      cdef int nx, ny
      cdef double slope, intercept, r_value, p_value, std_err, l, rms1, rms2, wmax, Z, E, sigma, T0_1, T0_2, sw1, sw2, d, wmean
      cdef np.ndarray[np.float64_t, ndim=1] moment = np.empty(5,dtype=np.float64)

      cdef np.ndarray[np.float64_t, ndim=1] s 
      cdef np.ndarray[np.float64_t, ndim=1] s_b 
      cdef np.ndarray[np.float64_t, ndim=1] k_back 
      cdef np.ndarray[np.float64_t, ndim=1] k
     
      #cdef double[::1] s
      cdef np.ndarray[np.float64_t, ndim=2] im 
      
      points = points - np.nanmin(points, axis=0)
       
      r = pysesa.lengthscale(points, res, lentype, taper, method)
      im = r.getdata()
      l = r.getlengthscale()
      self.lengthscale = l
      
      ny, nx= np.shape(im)

      # is all nans just return nans
      if np.sum(np.isnan(im)) == np.prod(np.shape(im)):

         self.data = [np.ones(24)*np.nan] 
         return

      else:

         #if proctype==1: #no spectral smoothing

         try:

            s, s_b, k, k_back = self._psd(im, nx, ny, nbin, res, proctype)
            slope, intercept, r_value, p_value, std_err, d = self._psparams(k_back, s_b, res)
            self.psdparams = [slope, intercept, r_value, p_value, std_err, d]
               
            # interpolate background spectrum onto wavenumber array
            s_b = np.interp(k,k_back,s_b)

            k = k[1:] #Nyquist frequency up
            s = s[1:]
            s_b = s_b[1:]

            # wavelengths and rms amplitudes
            wmax, wmean, rms1, rms2 = self._wav_rms(k, s, s_b, res)
            self.lengths = [wmax, wmean, rms1, rms2]
               
            # moments and moment parameters
            Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4 = self._moments(k, s_b, res)
            self.moments = [Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4]

            # concetanate and add effective slope
            #self.data = self.psdparams + [self.lengthscale] + self.lengths + self.moments + [np.arctan(sigma/l)/(np.pi/180)]
            self.data = self.psdparams + [self.lengthscale] + self.lengths + self.moments + [np.arctan(sigma/l)/(pi/180)]

         except:
                     
            self.data = [np.ones(24)*np.nan] 

         return

#         elif proctype==2: # with spectral smoothing

#            try:

#               s, s_b, k, k_back = self._psd(im, nx, ny, nbin, res, proctype)
#               slope, intercept, r_value, p_value, std_err, d = self._psparams(k_back, s_b, res)
#               self.psdparams = [slope, intercept, r_value, p_value, std_err, d]
#                              
#               # interpolate background spectrum onto wavenumber array
#               s_b = np.interp(k,k_back,s_b)

#               k = k[1:]
#               s = s[1:]
#               s_b = s_b[1:]

#               # wavelengths and rms amplitudes
#               wmax, wmean, rms1, rms2 = self._wav_rms(k, s, s_b, res)
#               self.lengths = [wmax, wmean, rms1, rms2]
#                              
#               # moments and moment parameters
#               Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4 = self._moments(k, s_b, res)
#               self.moments = [Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4]

#               # concetanate and add effective slope
#               #self.data = self.psdparams + [self.lengthscale] + self.lengths + self.moments + [np.arctan(sigma/l)/(np.pi/180)]
#               self.data = self.psdparams + [self.lengthscale] + self.lengths + self.moments + [np.arctan(sigma/l)/(pi/180)]

#            except:

#               self.data = [np.ones(24)*np.nan] 

#            return

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray _movingaverage(self, np.ndarray[np.float64_t, ndim=1] interval, int window_size):
      return np.convolve(interval, np.ones(int(window_size))/float(window_size), 'same')

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list _moments(self, np.ndarray[np.float64_t, ndim=1] k, np.ndarray[np.float64_t, ndim=1] s_b, double res):
      '''
      Return moments and moment parameters

      Syntax
      ----------
      Z, E, sigma, T0_1, T0_2, sw1, sw2, m0, m1, m2, m3, m4 = pysesa_spectral.spec._moments(k, s_b, res)

      Parameters
      ------------
      k : ndarray
   	   N x 1 wavenumbers
      s_b : ndarray
   	   N x 1 array of background power spectral density
      res : double
   	   spatial resolution of grid

      Returns
      ----------
      data: tuple of floats
        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

      '''
      cdef double Z, E, sigma, T0_1, T0_2, sw1, sw2
      cdef np.ndarray[np.float64_t, ndim=1] moment = np.empty(5,dtype=np.float64)
      cdef int i
      
      # get moments of spectrum
      for i from 0 <= i < 5:
      #for i in xrange(0,5):
         moment[i] = np.abs(trapz((k)**i,s_b)) #,np.median(np.gradient(k))))
         #moment[i] = abs(trapz((k)**i,s_b)) #,np.median(np.gradient(k))))

      # s is actually 10**length^4
      moment = np.sqrt((10**4)*moment)
      #moment = sqrt((10**4)*moment)

      # zero crossings per second
      Z = 2*(moment[2]/moment[0])**0.5
      #Z = 2*sqrt(moment[2]/moment[0])

      # extrema per second										
      E = 2*(moment[4]/moment[2])**0.5
      #E = 2*sqrt(moment[4]/moment[2]) 

      #rms
      sigma = (moment[2]/moment[0])

      # average period m0/m1
      T0_1 = (moment[0]/moment[1])
      #average period (m0/m2)^0.5
      T0_2 = T0_1**0.5  
      # spectral width parameter
      sw1 = (moment[0]*moment[2]/moment[1]**2-1)**0.5 
      # spectral width paramenter
      sw2 = np.abs(1 - moment[2]**2/(moment[0]*moment[4]))**0.5 
      #sw2 = abs(1 - moment[2]**2/(moment[0]*moment[4]))**0.5       

      return [Z, E, sigma, T0_1, T0_2, sw1, sw2] + moment.tolist()               
      
   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list _wav_rms(self, np.ndarray[np.float64_t, ndim=1] k, np.ndarray[np.float64_t, ndim=1] s, np.ndarray[np.float64_t, ndim=1] s_b, double res):
      '''
      Return max and mean wavelengths and rms amplitudes

      Syntax
      ----------
      wmax, wmean, rms1, rms2 = pysesa_spectral.spec._wav_rms(k, s, s_b, res)

      Parameters
      ----------
      k : ndarray
   	   N x 1 wavenumbers
      s : ndarray
   	   N x 1 array of power spectral density
      s_b : ndarray
   	   N x 1 array of background power spectral density
      res : double
   	   spatial resolution of grid

      Returns
      ----------
      data: tuple of floats
        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density

      '''
      cdef double wmax, wmean, rms1, rms2

      #cdef float pi = 3.14159265
      cdef np.ndarray[np.float64_t, ndim=1] ssb = np.empty(len(s),dtype=np.float64)
            
      ssb = self._movingaverage(s/s_b,len(s)/10)      
      
      # get peak wavelength
      wmax = res*(2*np.pi)/k[np.argmax(ssb)]       
      #wmax = res*(2*np.pi)/k[np.argmax(np.abs(s/s_b))] 
      #wmax = res*(2*pi)/k[np.argmax(abs(s/s_b))] 

      # mean wavelength
      wmean = res*(2*np.pi)/np.abs(trapz(s/s_b, k))
      #wmean = res*(2*pi)/abs(trapz(s/s_b, k))

      # get rms amplitudes
      rms1 = np.sqrt(np.abs(trapz(s, k)))/res
      rms2 = np.sqrt(np.abs(trapz(ssb, k)))/res
      #rms2 = np.sqrt(np.abs(trapz(s_b, k)))/res      
      #rms1 = sqrt(abs(trapz(s, k)))/res
      #rms2 = sqrt(abs(trapz(s_b, k)))/res
         
      return [wmax, wmean, rms1, rms2]               
                     
   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list _psd(self, np.ndarray[np.float64_t, ndim=2] im, int nx, int ny, int nbin, double res, int proctype):
      '''
      Return spectrum, wavenumber index, and background spectrum

      Syntax
      ----------
      s, s_b, k, k_back = pysesa_spectral.spec._psd(im, nx, ny, nbin, res, proctype)

      Parameters
      ----------
      im : ndarray
   	   2D array
      nx : int
   	   size of im in x dimension
      ny : int
   	   size of im in y dimension
      nbin : int
   	   number of bins for spectral binning
      res : double
   	   spatial resolution of grid
      proctype : int
   	   1 = no smoothing, 2 = smoothing

      Returns
      ----------
      data: tuple of ndarrays
   	s = power spectral density

        s_b = background power spectral density

        k = wavenumber index of s 

        k_back = wavenumber index of s_b

      '''
      from nifty import rg_space, field, about
      about.warnings='OFF'
                  
      cdef np.ndarray[np.float64_t, ndim=1] s 
      cdef np.ndarray[np.float64_t, ndim=1] s_b 
      cdef np.ndarray[np.float64_t, ndim=1] k_back 
      cdef np.ndarray[np.float64_t, ndim=1] k
      cdef object x_space, k_space

      #cdef double[::1] s
      #cdef double[::1] s_b
      #cdef double[::1] k_back
      #cdef double[::1] k
            
      # set up field
      if nx%2==0:
         x_space = rg_space((nx,ny), naxes=2)
      else:
         x_space = rg_space((nx-1,ny-1), naxes=2)
      k_space = x_space.get_codomain()

      if proctype==2:
         # and get the smoothed power spectrum
         if nx%2==0:
            s = field(x_space, target=k_space, val=im).power(smooth=1)
         else:
            s = field(x_space, target=k_space, val=im[:-1,:-1].T).power(smooth=1)
      else:
         # and get the power spectrum
         if nx%2==0:
            s = field(x_space, target=k_space, val=im).power(smooth=0) 
         else:
            s = field(x_space, target=k_space, val=im[:-1,:-1].T).power(smooth=0) 

      k = k_space.power_indices["kindex"]*res 
      # generate synthetic field and get binned power spectrum
      s_b = field(x_space, target=k_space, random="syn", spec=s).power(nbin=nbin) 
      k_back = k_space.power_indices["kindex"]*res 

      return [s, s_b, k, k_back]               
               
   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list _psparams(self, np.ndarray[np.float64_t, ndim=1] k_back, np.ndarray[np.float64_t, ndim=1] s_b, double res):
      '''
      Return slope, intercept, r_value, p_value, std_err, fractal dimension

      Syntax
      ----------
      slope, intercept, r_value, p_value, std_err, d = pysesa_spectral.spec._psdparams(k_back, s_b, res)

      Parameters
      ------------
      k_back : ndarray
   	   N x 1 wavenumbers
      s_b : ndarray
   	   N x 1 array of power spectral density
      res : double
   	   spatial resolution of grid

      Returns
      ----------
      data: tuple
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension

      '''
      cdef double slope, intercept, r_value, p_value, std_err, d
      
      slope, intercept, r_value, p_value, std_err = self._do_linreg(np.c_[np.log10(k_back),np.log10(s_b)])
      std_err = std_err*res

      ## give intercept square units 
      intercept = (10**intercept)**(1/(2-slope))
         
      if intercept>1:
         intercept=1
         slope=0

      if slope>0:
         slope=0
      
      # fractal dimension
      d = (8+slope)/2
      
      return [slope, intercept, r_value, p_value, std_err, d]

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef tuple _do_linreg(self, np.ndarray[np.float64_t, ndim=2] B):
      '''
      Do a robust linear regression

      Syntax
      ----------
      slope, intercept, r_value, p_value, std_err = pysesa_spectral.spec._do_linreg(B)

      Parameters
      ------------
      B : ndarray
   	   N x 2 array of wavenumber and power spectral density

      Returns
      ----------
      data: tuple
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

      '''

      cdef float slope, intercept, r_value, p_value, std_err
      cdef object regression

      # remove any rows with nans and infs
      B = B[np.where(np.logical_not(np.any(np.isinf(B),axis=1)))[0],:]
      B = B[np.where(np.logical_not(np.any(np.isnan(B),axis=1)))[0],:]

      # get OLS regression
      slope, intercept, r_value, p_value, std_err = linregress(B[:,0], B[:,1])
      # try to RLM regression for slope and intercept
      try:
         regression = smapi.RLM(B[:,1], smapi.add_constant(B[:,0], prepend=True) ).fit()
         intercept = regression.params[0]
         slope = regression.params[1]
      except:
         pass
      return (slope, intercept, r_value, p_value, std_err)

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list getdata(self):
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      data = pysesa.spectral.getdata()

      Parameters
      ------------
      self : instance
   	   pysesa.spectral instance

      Returns [requested through .getdata()]
      -----------------------------------------
      self.data: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        l = integral lengthscale

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density

        wmax = peak wavelength

        wmean = mean wavelength

        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        d = fractal dimension

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

        phi = effective slope (degrees)
      '''
      return self.data

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list getmoments(self):
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      moments = pysesa.spectral.getmoments()

      Parameters
      ------------
      self : instance
   	   pysesa.spectral instance

      Returns [requested through .getmoments()]
      ------------------------------------------
      self.moments: list
        Z = zero-crossings per unit length

        E = extreme per unit length

        sigma = RMS amplitude

        T0_1 = average spatial period (m_0/m_1)

        T0_2 = average spatial period (m_0/m_2)^0.5

        sw1 = spectral width 

        sw2 = spectral width (normalised radius of gyration)

        m0 = zeroth moment of spectrum

        m1 = first moment of spectrum

        m2 = second moment of spectrum

        m3 = third moment of spectrum

        m4 = fourth moment of spectrum

      '''
      return self.moments

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef double getlengthscale(self):
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      lengthscale = pysesa.spectral.getlengthscale()

      Parameters
      ------------
      self : instance
   	   pysesa.spectral instance

      Returns [requested through .getlengthscale()]
      -----------------------------------------------
      self.lengthscale: float
        l = integral lengthscale
      '''
      return self.lengthscale

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list getlengths(self):
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      lengths = pysesa.spectral.getlengths()

      Parameters
      ----------
      self : instance
   	   pysesa.spectral instance

      Returns [requested through .getlengths()]
      ------------------------------------------
      self.lengths: list
        wmax = peak wavelength

        wmean = mean wavelength

        rms1 = RMS amplitude from power spectral density

        rms2 = RMS amplitude from bin averaged power spectral density

      '''
      return self.lengths

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list getpsdparams(self):
      '''
      Calculate spectral statistics of a Nx3 point cloud

      Syntax
      ----------
      psdparams= pysesa.spectral.getstats()

      Parameters
      ------------
      self : instance
   	   pysesa.spectral.spec instance

      Returns [requested through .getpsdparams()]
      ---------------------------------------------
      self.psdparams: list
   	slope = slope of regression line through log-log 1D power spectral density

        intercept = intercept of regression line through log-log 1D power spectral density

        r_value = correlation of regression through log-log 1D power spectral density

        p_value = probability that slope of regression through log-log 1D power spectral density is not zero

        std_err = standard error of regression through log-log 1D power spectral density

        d = fractal dimension

      '''
      return self.psdparams




## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitoring & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

#"""
# ___      ___ ___ ___   _     _   _ 
#| _ \_  _/ __| __/ __| /_\   (_) (_)
#|  _/ || \__ \ _|\__ \/ _ \   _   _ 
#|_|  \_, |___/___|___/_/ \_\ (_) (_)
#     |__/                           

#+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
#|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
#+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
#|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
#+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+

#"""

import pysesa
import os
import shutil
import errno

__all__ = [
    'test',
    ]

def dircopy(src, dest):
    try:
        shutil.copytree(src, dest)
    except OSError as e:
        # If the error was caused because the source wasn't a directory
        if e.errno == errno.ENOTDIR:
            shutil.copy(src, dst)
        else:
            print('Directory not copied. Error: %s' % e)

def test():
   """
   PySESA - a Python framework for Spatially Explicit Spectral Analysis

   PySESA is an open-source project dedicated to provide a generic Python framework 
   for spatially explicit statistical analyses of point clouds and other geospatial data, 
   in the spatial and frequency domains, for use in the geosciences

   The program is detailed in:
   Buscombe, D. (2016) "Computational considerations for spatially explicit spectral analysis of point clouds and geospatial data", 86, 92-108, 10.1016/j.cageo.2015.10.004.

   :Author:  
       Daniel Buscombe
       Grand Canyon Monitoring and Research Center
       United States Geological Survey
       Flagstaff, AZ 86001
       dbuscombe@usgs.gov

   For more information visit http://dbuscombe-usgs.github.io/pysesa/

   :install:
       python setup.py install
       sudo python setup.py install
    
   :test:
       python -c "import pysesa; pysesa.test()"

   :license:
       GNU Lesser General Public License, Version 3
       (http://www.gnu.org/copyleft/lesser.html)
    
       This software is in the public domain because it contains materials that
       originally came from the United States Geological Survey, an agency of the
       United States Department of Interior. For more information, 
       see the official USGS copyright policy at
       http://www.usgs.gov/visual-id/credit_usgs.html#copyright
       Any use of trade, product, or firm names is for descriptive purposes only 
       and does not imply endorsement by the U.S. government.
    
   """
   # copy files over to somewhere read/writeable
   dircopy(pysesa.__path__[0], os.path.expanduser("~")+os.sep+'pysesa_test')
   shutil.copy(pysesa.__path__[0]+os.sep+'example_100000pts.xyz', os.path.expanduser("~")+os.sep+'pysesa_test'+os.sep+'example_100000pts.xyz')

   # general settings   
   infile = os.path.expanduser("~")+os.sep+'pysesa_test'+os.sep+'example_100000pts.xyz' 

   out = 0.5 #m output grid
   detrend = 4 #ODR plane
   proctype = 4 #Processing spatial and spectral parameters (no smoothing)
   mxpts = 2048 # max pts per window
   res = 0.05 #cm internal grid resolution
   nbin = 20 #number of bins for spectral binning
   lentype = 1 # l<0.5
   taper = 1 # Hann taper
   prc_overlap = 100 # 300% overlap between successive windows
   minpts = 32 # min pts per window

   pysesa.process(infile, out, detrend, proctype, mxpts, res, nbin, lentype, minpts, taper, prc_overlap)

if __name__ == '__main__':
   test()





## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

##"""
## ___      ___ ___ ___   _     _   _ 
##| _ \_  _/ __| __/ __| /_\   (_) (_)
##|  _/ || \__ \ _|\__ \/ _ \   _   _ 
##|_|  \_, |___/___|___/_/ \_\ (_) (_)
##     |__/                           
##                      __  _ __  _           
##    ____  ____ ______/ /_(_) /_(_)___  ____ 
##   / __ \/ __ `/ ___/ __/ / __/ / __ \/ __ \
##  / /_/ / /_/ / /  / /_/ / /_/ / /_/ / / / /
## / .___/\__,_/_/   \__/_/\__/_/\____/_/ /_/ 
##/_/                                         

##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+

##"""

# import libraries
from __future__ import division
import numpy as np
cimport numpy as np
cimport cython

try:
   from pykdtree.kdtree import KDTree
   pykdtree=1   
except:
   #print "install pykdtree for faster kd-tree operations: https://github.com/storpipfugl/pykdtree"
   from scipy.spatial import cKDTree as KDTree
   pykdtree=0   

# =========================================================
cdef class partition:

   '''
   Partition a Nx3 point cloud into M windows of nx3 points
   with specified spacing between centroids of adjacent windows
   and with specified overlap between windows.
   Implemented using a binary search tree for fast nearest neighbour 
   point check with boundary pruning

   Syntax
   ----------
   nr_pts = pysesa.partition(toproc, out, res, mxpts, minpts, prc_overlap).getdata()

   Parameters
   -----------
   toproc : ndarray
   	Nx3 point cloud

   Other Parameters
   -----------------
   out : float, *optional* [default = 0.5]
   	output grid resolution
   res : float, *optional* [default = 0.05]
   	spatial grid resolution to create a grid for the boundary pruning
   mxpts : float, *optional* [default = 1024]
   	maximum number of points allowed in a window
   minpts : float, *optional* [default = 16]
   	minimum number of points allowed in a window
   prc_overlap : float, *optional"  [default = 0]
        percentage overlap between windows

   Returns
   ----------
   self.data: list
   	list of M ndarrays, each containing n indices 
        of original point cloud, toproc, to partition space 
        to create M windows

   '''
   
   cdef object data, mytree, tree, tree2

   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   #==================================================
   def __init__(self, np.ndarray[np.float32_t, ndim=2] toproc, float out=0.5, float mxpts=1024, float minpts=16, float prc_overlap=0): #float res=0.05, int bp=0):

      '''
      Partition a Nx3 point cloud into M windows of nx3 points
      with specified spacing between centroids of adjacent windows
      and with specified overlap between windows.
      Implemented using a binary search tree for fast nearest neighbour 
      point check with optional boundary pruning

      Syntax
      ----------
      nr_pts = pysesa.partition(toproc, out, mxpts, minpts, prc_overlap).getdata()

      Parameters
      ----------
      toproc : ndarray
   	   Nx3 point cloud

      Other Parameters
      ----------
      out : float, *optional* [default = 0.5]
   	   output grid resolution
      mxpts : float, *optional* [default = 1024]
   	   maximum number of points allowed in a window
      minpts : float, *optional* [default = 16]
   	   minimum number of points allowed in a window
      prc_overlap : float, *optional"  [default = 0]
   	   percentage overlap between windows

      Returns
      ----------
      self.data: list
   	   list of M ndarrays, each containing n indices 
           of original point cloud, toproc, to partition space 
           to create M windows

      '''

      #get size of window
      cdef float win =  (out/2) * (1+(prc_overlap/100)) #np.multiply(out/2, 1+(prc_overlap/100))

      #remove any rows containing NaNs
      toproc = toproc[~np.isnan(toproc).any(axis=1)]

      # pre-allocate arrays     
      cdef float xmin = np.min(toproc[:,0])
      cdef float xmax = np.max(toproc[:,0])
      cdef float ymin = np.min(toproc[:,1])
      cdef float ymax = np.max(toproc[:,1])
      cdef int orig_pts = len(toproc)
      cdef int lenx = np.ceil((xmax-xmin)/out)
      cdef int leny = np.ceil((ymax-ymin)/out)
      cdef int lenx2 = lenx/2    
      cdef int leny2 = leny/2   
            
      cdef np.ndarray[np.float64_t, ndim=1] xvec = np.empty((lenx2,), dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=1] yvec = np.empty((leny2,), dtype=np.float64)   

#      cdef np.ndarray[np.float32_t, ndim=2] p = np.empty((lenx*leny,2), dtype=np.float32)   

      cdef np.ndarray[np.float64_t, ndim=2] xx = np.empty((leny,lenx), dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=2] yy = np.empty((leny,lenx), dtype=np.float64)

      # create output grid
      x = np.arange(xmin, xmax, out)
      y = np.arange(ymin, ymax, out)
      xx, yy = np.meshgrid(x, y)
      p = np.vstack([xx.flatten(),yy.flatten()]).transpose().astype('float32')
    
      # find all points within 'out' metres of each centroid in p 
      xvec = np.arange(xmin-2*out,xmax+2*out)
      yvec = np.arange(ymin-2*out,ymax+2*out)            

      # pre-allocate more arrays
      cdef int k
      #cdef tuple cx, cy
      cdef list indices_list
      #cdef list w = []
      #cdef list indices2 = []

      cdef np.ndarray[np.float64_t, ndim=2] xp = np.empty((len(yvec), len(xvec)), dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=2] yp = np.empty((len(yvec), len(xvec)), dtype=np.float64)
#      cdef np.ndarray[np.float32_t, ndim=1] dist2 = np.empty((len(xvec)*len(yvec),), dtype=np.float32)
# 
      mytree = KDTree(toproc[:,:2]) #, leafsize=len(toproc)/100)
      if pykdtree==1:
         complete=0
         while complete==0:
            try:
               dist, indices = mytree.query(p,mxpts, distance_upper_bound=win)
               if 'indices' in locals():
                  complete=1
            except:
               mxpts = np.max([1,mxpts-2])
               print "memory error, using %s max points" % (str(mxpts))         

      else:
         try:
            dist, indices = mytree.query(p,mxpts, distance_upper_bound=win, n_jobs=-1)
            #del p
         except:
            complete=0
            while complete==0:
               try:
                  dist, indices = mytree.query(p,mxpts, distance_upper_bound=win)
               except:
                  mxpts = np.max([1,mxpts-2])
                  print "memory error, using %s max points" % (str(mxpts))         

         finally:
            import dask.array as da
            complete=0
            while complete==0:
               try:
                  #dask implementation
                  dat = da.from_array(toproc[:,:2].astype('float64'), chunks=1000)
                  mytree = KDTree(dat, leafsize=len(dat)/100) # adding this leafsize option speeds up considerably
                  dbp = da.from_array(np.asarray(p), chunks=1000) 
                  #del p  
                  dist, indices = mytree.query(dbp.compute(),mxpts, distance_upper_bound=win) #.astype('float32')
                  del dbp
                  if 'indices' in locals():
                     complete=1
               except:
                  mxpts = np.max([1,mxpts-2])
                  print "memory error, using %s max points" % (str(mxpts))         


      # remove any indices associated with 'inf' distance
      indices = np.squeeze(indices[np.where(np.all(np.isinf(dist),axis=1) ==  False),:])
      dist = np.squeeze(dist[np.where(np.all(np.isinf(dist),axis=1) ==  False),:])

      # define null indices
      try:
         indices[indices==len(toproc)] = -999
      except:
         indices[indices==len(dat)] = -999  #dask implementation    

      indices_list = indices.tolist()
      del indices

      # remove null records
      #for k in xrange(len(indices_list)):
      for k from 0 <= k < len(indices_list):
         indices_list[k] = [x for x in indices_list[k] if x!=-999]

      #for k in xrange(len(indices_list)):
      for k from 0 <= k < len(indices_list):
         indices_list[k] = [x for x in indices_list[k] if x<len(toproc)]

      #try:
      #   # get the centroids
      #   #for k in xrange(len(indices_list)):
      #   for k from 0 <= k < len(indices_list):
      #      #w.append(np.mean([allpoints[i] for i in indices_list[k]], axis=0))
      #      w.append(np.mean([toproc[i,:2] for i in indices_list[k]], axis=0))

      #except:
      #   # get the centroids
      #   #for k in xrange(len(indices_list)):
      #   for k from 0 <= k < len(indices_list):
      #      w.append(dat[indices_list[k],:2].mean(axis=0).compute()) #dask implementation

      #cx,cy = zip(*w)
      #del w

      #yp, xp = np.meshgrid(yvec, xvec)

      # Use KDTree to answer the question: "which point of set (x,y) is the
      # nearest neighbors of those in (xp, yp)"

      #if pykdtree==1:
      #   dist2, _ = mytree.query(np.c_[xp.ravel(), yp.ravel()].astype('float32'), k=1)
      #else:
      #   try:      
      #      dist2, _ = mytree.query(np.c_[xp.ravel(), yp.ravel()].astype('float32'), k=1, n_jobs=-1)
      #   except:
      #      dist2, _ = mytree.query(np.c_[xp.ravel(), yp.ravel()].astype('float32'), k=1)      

      #if bp==1: #boundary pruning
      #   # Select points sufficiently far away (use hypoteneuse of the triangle made by res and res)
      #   tree2 = KDTree(np.c_[xp.ravel()[(dist2 > np.hypot(out, out))], yp.ravel()[(dist2 > np.hypot(out, out))]])

      #   if pykdtree==1:     
      #      dist3, _ = tree2.query(np.c_[cx,cy], distance_upper_bound=win) #distance_upper_bound=out)
      #   else: 
      #      try:
      #         dist3, _ = tree2.query(np.c_[cx,cy], distance_upper_bound=win, n_jobs=-1) #distance_upper_bound=out)
      #      except:
      #         dist3, _ = tree2.query(np.c_[cx,cy], distance_upper_bound=win)      
      
      #   m2 = np.where(dist3 < np.hypot(out, out))[0]

      #else:
      #   m2 = np.where(dist2 < np.hypot(out, out))[0]

      #m2 = m2[np.where(m2<len(indices_list))[0]]

      ## do the pruning
      ##for k in xrange(len(m2)):
      #for k from 0 <= k < len(m2):
      #   if len(indices_list[m2[k]])>minpts:
      #      indices2.append(indices_list[m2[k]])

      #if len(indices2)<1:
      #   print "no returned windows: either increase 'out' or 'prc_overlap'"

      self.data = indices_list #indices2

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef list getdata(self):
      '''
      Partition a Nx3 point cloud into M windows of nx3 points
      with specified spacing between centroids of adjacent windows
      and with specified overlap between windows.
      Implemented using a binary search tree for fast nearest neighbour 
      point check with boundary pruning

      Syntax
      ----------
      nr_pts = pysesa.partition.getdata()

      Parameters
      ----------
      self : instance
   	   pysesa.partition instance

      Returns
      ----------
      self.data: list
   	   list of M ndarrays, each containing n indices 
           of original point cloud, toproc, to partition space 
           to create M windows
      '''
      return self.data


      #cdef np.ndarray[np.float64_t, ndim=2] indices = np.empty((len(p),mxpts), dtype=np.int64)
       #cdef np.ndarray[np.float64_t, ndim=2] xx = np.empty((int(np.ceil((ymax-ymin)/out)),int(np.ceil((xmax-xmin)/out))), dtype=np.float64)
      #cdef np.ndarray[np.float64_t, ndim=2] yy = np.empty((int(np.ceil((ymax-ymin)/out)),int(np.ceil((xmax-xmin)/out))), dtype=np.float64)
      #dbp = db.from_sequence(p, npartitions = 1000) #dask bag
      #del p #dask

      # format points for kd-tree
      #allpoints = zip(toproc[:,0].ravel(), toproc[:,1].ravel())
      #del toproc
       
      # get the tree for the entire point cloud
      #mytree = cKDTree(allpoints) 
      
      # largest inscribed square has side length = sqrt(2)*radius
      #dist, indices = mytree.query(p,mxpts, distance_upper_bound=win)

      #dist, indices = mytree.query(dbp.compute(),mxpts, distance_upper_bound=win) #dask implementation 1
      
      #cdef int bp=0 #boundary prining flag
      
      # no idea why this doesnt work
      #cdef np.ndarray[np.float64_t, ndim=1] x = np.empty(lenx, dtype=np.float64)
      #cdef np.ndarray[np.float64_t, ndim=1] y = np.empty(leny, dtype=np.float64)
      #try:
      #   tree = KDTree(allpoints, leafsize=len(allpoints)/100)
      #except:
      #   tree = KDTree(dat, leafsize=len(dat)/100) #dask implementation 2 # leafsize=len(dat)
      #   del dat
      #p = list(np.vstack([xx.flatten(),yy.flatten()]).transpose()) 
      #cdef list p
      #cdef list allpoints 
#from scipy.spatial import cKDTree
#import dask.bag as db
   
#data_type = np.float64
#ctypedef np.float64_t data_type_f64t

#ctypedef double (*metric_ptr)(double[:, ::1], np.intp_t, np.intp_t)
            



#!/usr/bin/env python
# -*- coding: utf-8 -*-
## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
PySESA - a Python framework for Spatially Explicit Spectral Analysis

PySESA is an open-source project dedicated to provide a generic Python framework 
for spatially explicit statistical analyses of point clouds and other geospatial data, 
in the spatial and frequency domains, for use in the geosciences

The program is detailed in:
   Buscombe, D. (2016) "Computational considerations for spatially explicit spectral analysis of point clouds and geospatial data", 86, 92-108, 10.1016/j.cageo.2015.10.004.

:Author:  
    Daniel Buscombe
    Grand Canyon Monitoring and Research Center
    United States Geological Survey
    Flagstaff, AZ 86001
    dbuscombe@usgs.gov

For more information visit http://dbuscombe-usgs.github.io/pysesa/

:install:
    pip install pysesa
    python setup.py install
    sudo python setup.py install
    
:test:
    python -c "import pysesa; pysesa.test()"

:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
    
    This software is in the public domain because it contains materials that
    originally came from the United States Geological Survey, an agency of the
    United States Department of Interior. For more information, 
    see the official USGS copyright policy at
    http://www.usgs.gov/visual-id/credit_usgs.html#copyright
    Any use of trade, product, or firm names is for descriptive purposes only 
    and does not imply endorsement by the U.S. government.
    
"""
#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import os, sys, glob
import inspect
try:
    import numpy as np
except:
    msg = ("No module named numpy. "
           "Please install numpy first, it is needed before installing PySESA.")
    raise ImportError(msg)

from distutils.core import setup
from distutils.extension import Extension
#from setuptools import setup, Extension

# Directory of the current file 
SETUP_DIRECTORY = os.path.dirname(os.path.abspath(inspect.getfile(
    inspect.currentframe())))

# Set this to True to enable building extensions using Cython.
# Set it to False to build extensions from the C file (that
# was previously created using Cython).
# Set it to 'auto' to build with Cython if available, otherwise
# from the C file.
USE_CYTHON = True

if USE_CYTHON:
   try:
      from Cython.Distutils import build_ext
      from Cython.Build import cythonize
   except:
      msg = ("No module named Cython. "
           "Please install Cython first, it is needed before installing pysesa.")
      raise ImportError(msg)

# Read version from distmesh/__init__.py
with open(os.path.join('pysesa', '__init__.py')) as f:
    line = f.readline()
    while not line.startswith('__version__'):
        line = f.readline()
exec(line, globals())

ext_modules = [ ]
cmdclass = { }

#ext_modules += cythonize(Extension('_loadtxt',sources=['pysesa/loadtxt.pyx', 'pysesa/loadtxt.cpp'], language="c++"))

if USE_CYTHON:
    ext_modules += [
        Extension("pysesa.plot", [ "pysesa/plot.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.read", [ "pysesa/_read.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.write", [ "pysesa/_write.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.detrend", [ "pysesa/detrend.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.lengthscale", [ "pysesa/lengthscale.pyx" ],
        include_dirs=[np.get_include()]),        
        Extension("pysesa.partition", [ "pysesa/partition.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.spectral", [ "pysesa/spectral.pyx" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.spatial", [ "pysesa/_spatial.pyx" ],
        include_dirs=[np.get_include()]),    
        Extension("pysesa.sgolay", [ "pysesa/_sgolay.pyx" ],
        include_dirs=[np.get_include()]),
        Extension('_RunningStats',sources=['pysesa/RunningStats_wrap.cxx', 'pysesa/RunningStats.cpp']),
        ]
    cmdclass.update({ 'build_ext': build_ext })
else:
    ext_modules += [
        Extension("pysesa.plot", [ "pysesa/plot.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.read", [ "pysesa/_read.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.write", [ "pysesa/_write.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.detrend", [ "pysesa/detrend.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.lengthscale", [ "pysesa/lengthscale.c" ],
        include_dirs=[np.get_include()]), 
        Extension("pysesa.partition", [ "pysesa/partition.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.spectral", [ "pysesa/spectral.c" ],
        include_dirs=[np.get_include()]),
        Extension("pysesa.spatial", [ "pysesa/_spatial.c" ],
        include_dirs=[np.get_include()]),    
        Extension("pysesa.sgolay", [ "pysesa/_sgolay.c" ],
        include_dirs=[np.get_include()]),
        Extension('_RunningStats',sources=['pysesa/RunningStats_wrap.cxx', 'pysesa/RunningStats.cpp']),
    ]
install_requires = [
    'numpy','scipy','matplotlib', 'cython', 'statsmodels', 'ift_nifty', 'joblib','dask', 'toolz', 'dill'
]

def setupPackage():
   setup(name='pysesa',
         version=__version__,
         description='PySESA is an open-source project dedicated to provide a generic Python framework for spatially explicit statistical analyses of point clouds and other geospatial data, in the spatial and frequency domains, for use in the geosciences. The program is detailed in  Buscombe, D. (2016) "Computational considerations for spatially explicit spectral analysis of point clouds and geospatial data", 86, 92-108, 10.1016/j.cageo.2015.10.004.',
         #long_description=long_description,
         classifiers=[
             'Intended Audience :: Science/Research',
             'Intended Audience :: Developers',
             'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
             'Programming Language :: Python',
             'Programming Language :: Python :: 2.7',
             'Programming Language :: Cython',
             'Topic :: Scientific/Engineering',
         ],
         keywords='point clouds',
         author='Daniel Buscombe',
         author_email='dbuscombe@usgs.gov',
         url='https://github.com/dbuscombe-usgs/pysesa',
         download_url ='https://github.com/dbuscombe-usgs/pysesa/archive/master.zip',
         install_requires=install_requires,
         license = "GNU GENERAL PUBLIC LICENSE v3",
         packages=['pysesa'],
         cmdclass = cmdclass,
         ext_modules=ext_modules,
         platforms='OS Independent',
         package_data={'pysesa': ['*.xyz','*.h']}
   )

if __name__ == '__main__':
    # clean --all does not remove extensions automatically
    if 'clean' in sys.argv and '--all' in sys.argv:
        import shutil
        # delete complete build directory
        path = os.path.join(SETUP_DIRECTORY, 'build')
        try:
            shutil.rmtree(path)
        except:
            pass
        # delete all shared libs from lib directory
        path = os.path.join(SETUP_DIRECTORY, 'pysesa')
        for filename in glob.glob(path + os.sep + '*.pyd'):
            try:
                os.remove(filename)
            except:
                pass
        for filename in glob.glob(path + os.sep + '*.so'):
            try:
                os.remove(filename)
            except:
                pass
    setupPackage()




## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

##"""                       
## ___      ___ ___ ___   _     _   _ 
##| _ \_  _/ __| __/ __| /_\   (_) (_)
##|  _/ || \__ \ _|\__ \/ _ \   _   _ 
##|_|  \_, |___/___|___/_/ \_\ (_) (_)
##     |__/                           
##    __                 __  __                    __   
##   / /__  ____  ____ _/ /_/ /_  ______________ _/ /__ 
##  / / _ \/ __ \/ __ `/ __/ __ \/ ___/ ___/ __ `/ / _ \
## / /  __/ / / / /_/ / /_/ / / (__  ) /__/ /_/ / /  __/
##/_/\___/_/ /_/\__, /\__/_/ /_/____/\___/\__,_/_/\___/ 
##             /____/                                   

##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##"""

# import libraries
from __future__ import division
import numpy as np
cimport numpy as np
cimport cython
#from scipy.interpolate import griddata
from scipy.integrate import trapz
#from scipy.spatial import cKDTree as KDTree

#try:
#   from pykdtree.kdtree import KDTree
#   pykdtree=1   
#except:
#   #print "install pykdtree for faster kd-tree operations: https://github.com/storpipfugl/pykdtree"
#   from scipy.spatial import cKDTree as KDTree
#   pykdtree=0   

from scipy.spatial import cKDTree as KDTree

import RunningStats

# suppress divide and invalid warnings
np.seterr(divide='ignore')
np.seterr(invalid='ignore')
np.seterr(over='ignore')
np.seterr(under='ignore')
import warnings
warnings.filterwarnings("ignore")

# =========================================================
cdef class lengthscale:
   '''
   Calculates the integral lengthscale of a Nx3 point cloud
   using 1 of 3 available methods
   and also returns the tapered 2D grid of 3D pointcloud for spectral analysis

   Syntax
   ----------
   im = pysesa.lengthscale(points, res, lentype, taper, method).getdata()
   lengthscale = pysesa.lengthscale(points, res, lentype, taper, method).getlengthscale()

   Parameters
   ------------
   points : ndarray
   	Nx3 point cloud

   Other Parameters
   ------------------
   res : float, *optional* [default = 0.05]
        spatial grid resolution to create a grid
   lentype : int, *optional* [default = 0, l<0.5]
   	lengthscale type:
        1, l<0.5

        2, l<1/e

        3, l<0

   taper : int, *optional* [default = Hanning]
   	flag for taper type:
        1, Hanning (Hann)

        2, Hamming

        3, Blackman

        4, Bartlett

   method : str, *optional* [default = 'nearest']
   	gridding type

   Returns [requested through .getdata()]
   -----------------------------------------
   self.data: ndarray
   	tapered 2D grid of 3D pointcloud

   Returns [requested through .getlengthscale()]
   ----------------------------------------------
   self.lengthscale: float
   	integral lengthscale

   '''

   cdef object data, lengthscale, tree, rs1

   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   # =========================================================
   def __init__(self, np.ndarray[np.float64_t, ndim=2] points, float res=0.05, int lentype=1, int taper=1, str method='nearest'): 
      '''
      Calculates the integral lengthscale of a Nx3 point cloud
      using 1 of 3 available methods
      and also returns the tapered 2D grid of 3D pointcloud for spectral analysis

      Syntax
      ----------
      im = pysesa.lengthscale(points, res, lentype, taper, method).getdata()
      lengthscale = pysesa.lengthscale(points, res, lentype, taper, method).getlengthscale()

      Parameters
      ----------
      points : ndarray
   	   Nx3 point cloud

      Other Parameters
      ----------
      res : float, *optional* [default = 0.05]
           spatial grid resolution to create a grid
      lentype : int, *optional* [default = 1, l<0.5]
   	   lengthscale type:
           1, l<0.5
           2, l<1/e
           3, l<0
      taper : int, *optional* [default = Hanning]
   	   flag for taper type:
           1, Hanning (Hann)
           2, Hamming
           3, Blackman
           4, Bartlett
      method : str, *optional* [default = 'nearest']
   	   gridding type

      Returns [requested through .getdata()]
      ----------
      self.data: ndarray
   	   tapered 2D grid of 3D pointcloud

      Returns [requested through .getlengthscale()]
      ----------
      self.lengthscale: float
   	   integral lengthscale

      '''

      points = points - np.nanmin(points, axis=0)
      
      # pre-allocate some more arrays and get the ranges of x and y
      cdef float nnx = np.max(points[:,0]) - np.min(points[:,0])
      cdef float nny = np.max(points[:,1]) - np.min(points[:,1])
      cdef float lenx
      cdef int nx, ny, i

      # enfore square matrix in gridding
      if nny!=nnx:
         lenx = np.ceil(nnx/(res*(nnx/nny)))
      else:
         lenx = np.ceil(nnx/res)

      cdef np.ndarray[np.float64_t, ndim=2] grid_x = np.empty((lenx,lenx), dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=2] grid_y = np.empty((lenx,lenx), dtype=np.float64)

      cdef np.ndarray[np.float64_t, ndim=2] im = np.empty((lenx,lenx), dtype=np.float64)

      if nnx==0 or nny==0:
         self.data = np.nan
         self.lengthscale = 0.0

      else:

         # do the gridding
         if nny>nnx:
            grid_x, grid_y = np.meshgrid( np.arange(np.min(points,axis=0)[0], np.max(points,axis=0)[0], res*(nnx/nny)), np.arange(np.min(points,axis=0)[1], np.max(points,axis=0)[1], res) )      
         elif nny<nnx:
            grid_x, grid_y = np.meshgrid( np.arange(np.min(points,axis=0)[0], np.max(points,axis=0)[0], res), np.arange(np.min(points,axis=0)[1], np.max(points,axis=0)[1], res*(nny/nnx)) )      
         else:
            grid_x, grid_y = np.meshgrid( np.arange(np.min(points,axis=0)[0], np.max(points,axis=0)[0], res), np.arange(np.min(points,axis=0)[1], np.max(points,axis=0)[1], res) )  

         #im = griddata(points[:,:2], points[:,2], (grid_x, grid_y), method=method)

         #tree = KDTree(zip(points[:,0], points[:,1]))
         tree = KDTree(zip(points[:,0], points[:,1])) #, leafsize=1000)
         #tree = KDTree(points[:,:2])

         #if pykdtree==1:
         #   _, inds = tree.query(np.c_[grid_x.flatten(),grid_y.flatten()].astype('float32'), k = 1)
         #else:
         #   _, inds = tree.query(zip(grid_x.flatten(), grid_y.flatten()), k = 1)

         _, inds = tree.query(zip(grid_x.flatten(), grid_y.flatten()), k = 1)

         im = points[:,2].flatten()[inds].reshape(np.shape(grid_x))

         #rs1 = RunningStats.RunningStats()
         #for k in im.flatten():
         #   rs1.Push(k)

         #im = im - rs1.Mean() #np.mean(im)
         ny, nx= np.shape(im)

         #taper
         im = self._taper_im(im, taper)
         self.data = im

         #get integral lengthscale
         out = self._get_lengthscale(im, lentype, res, nx, ny)
         self.lengthscale = out


   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray _radial_data(self, np.ndarray[np.float64_t, ndim=2] data, int annulus_width=1):
       '''
       Efficient radial average of matrix

       Syntax
       ----------
       h = pysesa.lengthscale._radial_data(data, annulus_width)

       Parameters
       ----------
       data : ndarray
   	   2D grid
       annulus_width : int [default = 1]
           angular resolution of sector mean

       Returns
       ----------
       radialdatamean: ndarray
   	   radial average 1D array

       '''
       cdef int npix, npiy, nrad, irad
       cdef double rmax

       npix, npiy = np.shape(data)

       cdef np.ndarray[np.float64_t, ndim=2] r = np.empty((npix, npiy),dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=2] x = np.empty((npix, npiy),dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=2] y = np.empty((npix, npiy),dtype=np.float64)  
       cdef np.ndarray[np.float64_t, ndim=1] x1 = np.empty(npix,dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=1] y1 = np.empty(npiy,dtype=np.float64)

       cdef np.ndarray[np.float64_t, ndim=1] minrad = np.empty(1,dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=1] maxrad = np.empty(1,dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=1] dr = np.empty(1,dtype=np.float64)

       x1 = np.arange(-np.float64(npix/2),np.float64(npix/2))
       y1 = np.arange(-np.float64(npiy/2),np.float64(npiy/2))
       x,y = np.meshgrid(y1,x1)
       r = np.abs(x+1j*y)

       rmax = np.max(r)
       dr = np.abs([x[0,0] - x[0,1]]) * annulus_width
    
       cdef int sizeout = np.ceil(rmax/dr)

       cdef np.ndarray[np.float64_t, ndim=1] radial = np.empty(sizeout,dtype=np.float64)
       cdef np.ndarray[np.float64_t, ndim=1] radialdatamean = np.empty(sizeout,dtype=np.float64)

       radial = np.arange(rmax/dr)*dr + np.float64(dr/2)
       nrad = len(radial)

       # Loop through the bins
       for irad from 0 <= irad < nrad:
         minrad = irad*dr
         maxrad = minrad + dr
         radialdatamean[irad] = data[(r>=minrad) * (r<maxrad)].mean()

       return radialdatamean

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef double _get_lengthscale(self, np.ndarray[np.float64_t, ndim=2] im, int lentype, double res, int nx, int ny):
      '''
      Calculates the integral lengthscale of 
      a tapered 2D grid of a Nx3 point cloud
      using 1 of 3 available methods

      Syntax
      ----------
      lengthscale = pysesa.lengthscale._get_lengthscale(points, lentype, res, nx, ny)

      Parameters
      ----------
      im : ndarray
   	   tapered 2D grid of Nx3 point cloud
      res : float
           spatial grid resolution to create a grid
      lentype : int
   	   lengthscale type:
           1, l<0.5
           2, l<1/e
           3, l<0
      nx : int
      	   size of im in x direction
      ny : int
      	   size of im in y direction

      Returns
      ----------
      self.lengthscale: float
   	   integral lengthscale

      '''
      cdef double ma
      cdef np.ndarray[np.complex128_t, ndim=2] ft1 = np.empty((nx,ny),dtype=np.complex128)
      cdef np.ndarray[np.float64_t, ndim=2] mag1 = np.empty((nx,ny),dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=2] auto = np.empty((nx,ny),dtype=np.float64)
      cdef np.ndarray[np.float64_t, ndim=2] autoarray = np.empty((nx,ny),dtype=np.float64)

      # =========================================================
      # autocorrelation
      # 2D fourier transform of demeaned image
      if np.sum(np.isnan(im)) == np.prod(np.shape(im)):
         return np.nan

      if np.prod(np.shape(im))<2:
         return np.nan

      ft1 = np.fft.fft2(im)
      mag1 = pow(abs(ft1),2) # power spectrum
      # autocovariance as inverse fourier transform as zero-centred power spectrum
      auto = np.fft.fftshift(np.real(np.fft.ifft2(mag1)))
      autoarray = np.asarray(auto) # make 1d to find max
      ma = autoarray.max() # find max
      # get integral lengthscale
      if ma>0:
         auto = np.dot(auto,pow(ma,-1))
         h = self._radial_data(np.squeeze(auto))

         try:
            if lentype==1: #l<0.5
               return res*(2*np.pi)*np.abs(trapz(h[:np.where(h<0.5)[0][0]+1], np.arange(np.where(h<0.5)[0][0]+1) ))
            elif lentype==2: #l<1/e
               return res*(2*np.pi)*np.abs(trapz(h[:np.where(h<(1/np.exp(1)))[0][0]+1], np.arange(np.where(h<(1/np.exp(1)))[0][0]+1) ))
            else: #l<0
               return res*(2*np.pi)*np.abs(trapz(h[:np.where(h<0)[0][0]+1], np.arange(np.where(h<0)[0][0]+1) ))
         except:
            return np.nan
      else:
         return np.nan

#         try:
#            if lentype==1: #l<0.5
#               return res*((2*np.pi)*(np.where(h<0.5)[0][0]+1))
#            elif lentype==2: #l<1/e
#               return res*((2*np.pi)*(np.where(h<(1/np.exp(1)))[0][0]+1))
#            else: #l<0
#               return res*(np.where(h<0)[0][0]+1)
#         except:
#            return np.nan
#      else:
#         return np.nan

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] _Hanning2D(self, np.ndarray[np.float64_t, ndim=2] im):
      '''
      Returns a 2D Hanning (a weighted cosine) tapered 2D data

      Syntax
      ----------
      im = pysesa.lengthscale._Hanning2D(im)

      Parameters
      ----------
      im : ndarray
   	   2D grid of Nx3 point cloud

      Returns
      ----------
      im: ndarray
   	   tapered 2D grid 

      '''
      cdef int nx, ny
      ny, nx= np.shape(im)

      cdef np.ndarray[np.float64_t, ndim=2] Wss = np.empty((ny, nx),dtype=np.float64)

      Wss = np.sqrt(np.outer(np.hanning(ny),np.hanning(nx)))
      return im*Wss

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] _Hamming2D(self, np.ndarray[np.float64_t, ndim=2] im):
      '''
      Returns a 2D Hamming (a weighted cosine) tapered 2D data

      Syntax
      ----------
      im = pysesa.lengthscale._Hamming2D(im)

      Parameters
      ----------
      im : ndarray
   	   2D grid of Nx3 point cloud

      Returns
      ----------
      im: ndarray
   	   tapered 2D grid 

      '''
      cdef int nx, ny
      ny, nx= np.shape(im)

      cdef np.ndarray[np.float64_t, ndim=2] Wss = np.empty((ny, nx),dtype=np.float64)

      Wss = np.sqrt(np.outer(np.hamming(ny),np.hamming(nx)))
      return im*Wss

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] _Blackman2D(self, np.ndarray[np.float64_t, ndim=2] im):
      '''
      Returns a 2D Blackman (a summation of cosines) tapered 2D data

      Syntax
      ----------
      im = pysesa.lengthscale._Blackman2D(im)

      Parameters
      ----------
      im : ndarray
   	   2D grid of Nx3 point cloud

      Returns
      ----------
      im: ndarray
   	   tapered 2D grid 

      '''
      cdef int nx, ny
      ny, nx= np.shape(im)

      cdef np.ndarray[np.float64_t, ndim=2] Wss = np.empty((ny, nx),dtype=np.float64)

      Wss = np.sqrt(np.outer(np.blackman(ny),np.blackman(nx)))
      return im*Wss

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] _Bartlett2D(self, np.ndarray[np.float64_t, ndim=2] im):
      '''
      Returns a 2D Bartlett (a triangular) tapered 2D data

      Syntax
      ----------
      im = pysesa.lengthscale._Bartlett2D(im)

      Parameters
      ----------
      im : ndarray
   	   2D grid of Nx3 point cloud

      Returns
      ----------
      im: ndarray
   	   tapered 2D grid 

      '''
      cdef int nx, ny
      ny, nx= np.shape(im)

      cdef np.ndarray[np.float64_t, ndim=2] Wss = np.empty((ny, nx),dtype=np.float64)

      Wss = np.sqrt(np.outer(np.bartlett(ny),np.bartlett(nx)))
      return im*Wss

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] _taper_im(self, np.ndarray[np.float64_t, ndim=2] im, int taper):
      '''
      Returns the tapered 2D grid of 3D pointcloud for spectral analysis

      Syntax
      ----------
      im = pysesa.lengthscale._taper(im, taper)

      Parameters
      ----------
      im : ndarray
   	   2D grid of Nx3 point cloud
      taper : int
   	   flag for taper type:
           1, Hanning (Hann)
           2, Hamming
           3, Blackman
           4, Bartlett

      Returns
      ----------
      im: ndarray
   	   tapered 2D grid of 3D pointcloud

      '''

      rs1 = RunningStats.RunningStats()
      for k in im.flatten():
         rs1.Push(k)

      im[np.isnan(im)] = rs1.Mean() #np.mean(im.flatten())
      im[np.isinf(im)] = rs1.Mean() #np.mean(im.flatten())
      im[im<=0] = rs1.Mean() #np.mean(im.flatten())

      if taper==1:
         im = self._Hanning2D(im)
      elif taper==2:
         im = self._Hamming2D(im)
      elif taper==3:
         im = self._Blackman2D(im)
      else:
         im = self._Bartlett2D(im)
      return im

   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef double getlengthscale(self):
      '''
      Returns the integral lengthscale

      Syntax
      ----------
      lengthscale = pysesa.lengthscale.getlengthscale()

      Parameters
      ----------
      self : instance
   	   pysesa.lengthscale instance

      Returns
      ----------
      self.lengthscale: float
   	   integral lengthscale

      '''
      return self.lengthscale


   # =========================================================
   @cython.boundscheck(False)
   @cython.cdivision(True)
   @cython.wraparound(False)
   @cython.nonecheck(False)
   cpdef np.ndarray[np.float64_t, ndim=2] getdata(self):
      '''
      Returns the tapered 2D grid of 3D pointcloud for spectral analysis

      Syntax
      ----------
      im = pysesa.lengthscale.getdata()

      Parameters
      ----------
      self : instance
   	   pysesa.lengthscale instance

      Returns 
      ----------
      self.data: ndarray
   	   tapered 2D grid of 3D pointcloud

      '''
      return self.data





## PySESA (Python program for Spatially Explicit Spectral Analysis) 
## has been developed at the Grand Canyon Monitorinf & Research Center,
## U.S. Geological Survey
##
## Author: Daniel Buscombe
## Project homepage: <https://github.com/dbuscombe-usgs/pysesa>
##
##This software is in the public domain because it contains materials that originally came from 
##the United States Geological Survey, an agency of the United States Department of Interior. 
##For more information, see the official USGS copyright policy at 
##http://www.usgs.gov/visual-id/credit_usgs.html#copyright
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program. If not, see <http://www.gnu.org/licenses/>.

##"""
## ___      ___ ___ ___   _     _   _ 
##| _ \_  _/ __| __/ __| /_\   (_) (_)
##|  _/ || \__ \ _|\__ \/ _ \   _   _ 
##|_|  \_, |___/___|___/_/ \_\ (_) (_)
##     |__/                           
##   _____             _ __       __               ______      __           
##  / ___/____ __   __(_) /______/ /____  __      / ____/___  / /___ ___  __
##  \__ \/ __ `/ | / / / __/ ___/ //_/ / / /_____/ / __/ __ \/ / __ `/ / / /
## ___/ / /_/ /| |/ / / /_(__  ) ,< / /_/ /_____/ /_/ / /_/ / / /_/ / /_/ / 
##/____/\__,_/ |___/_/\__/____/_/|_|\__, /      \____/\____/_/\__,_/\__, /  
##                                 /____/                          /____/   

##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##|b|y| |D|a|n|i|e|l| |B|u|s|c|o|m|b|e|
##+-+-+ +-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##|d|b|u|s|c|o|m|b|e|@|u|s|g|s|.|g|o|v|
##+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##|U|.|S|.| |G|e|o|l|o|g|i|c|a|l| |S|u|r|v|e|y|
##+-+-+-+-+ +-+-+-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+
##"""

# import libraries
from __future__ import division
import numpy as np
cimport numpy as np
cimport cython
import scipy.signal as sp

# =========================================================
cdef class sgolay:
           
    '''
    Create a Savitsky-Golay digital filter from a 2D signal
    based on code from http://www.scipy.org/Cookbook/SavitzkyGolay

    Syntax
    ----------
    Z = pysesa.sgolay(z, window_size, order).getdata()

    Parameters
    ----------
    z : array_like, shape (N,)
      the 2D signal.
    window_size : int
       the length of the window. Must be an odd integer number.
    order : int
       the order of the polynomial used in the filtering.
       Must be less than `window_size` - 1.

    Returns
    -------
    self.data : ndarray, shape (N)
       the smoothed signal.

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    '''
 
    cdef object data
    
    # =========================================================
    @cython.boundscheck(False)
    @cython.cdivision(True)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    def __init__(self, np.ndarray[np.float32_t, ndim=2] z, int window_size, int order ):
       '''
       Create a Savitsky-Golay digital filter from a 2D signal
       based on code from http://www.scipy.org/Cookbook/SavitzkyGolay

       Syntax
       ----------
       Z = pysesa.sgolay(z, window_size, order).getdata()

       Parameters
       ----------
       z : array_like, shape (N,)
         the 2D signal.
       window_size : int
          the length of the window. Must be an odd integer number.
       order : int
          the order of the polynomial used in the filtering.
          Must be less than `window_size` - 1.

       Returns
       -------
       self.data : ndarray, shape (N)
          the smoothed signal.

       References
       ----------
       .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
          Data by Simplified Least Squares Procedures. Analytical
          Chemistry, 1964, 36 (8), pp 1627-1639.
       .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
          W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
          Cambridge University Press ISBN-13: 9780521880688
       '''

       # number of terms in the polynomial expression
       cdef float n_terms = ( order + 1 ) * ( order + 2)  / 2.0

       if  window_size % 2 == 0:
           #raise ValueError('window_size must be odd')
           window_size = window_size + 1

       if window_size**2 < n_terms:
           raise ValueError('order is too high for the window size')

       cdef int half_size = window_size // 2

       # exponents of the polynomial. 
       # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ... 
       # this line gives a list of two item tuple. Each tuple contains 
       # the exponents of the k-th term. First element of tuple is for x
       # second element for y.
       # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
       cdef list exps = [] 
       exps = [ (k-n, n) for k in xrange(order+1) for n in xrange(k+1) ]

       # coordinates of points
       cdef np.ndarray[np.float64_t,ndim=1] ind = np.arange(-half_size, half_size+1, dtype=np.float64)
       cdef np.ndarray[np.float64_t,ndim=1] dx = np.empty(len(ind)*window_size, dtype=np.float64) 
       dx = np.repeat( ind, window_size )
       cdef np.ndarray[np.float64_t,ndim=1] dy = np.empty(len(ind)*window_size, dtype=np.float64) 
       dy = np.tile( ind, [window_size, 1]).reshape(window_size**2, )

       # build matrix of system of equation
       cdef np.ndarray[np.float64_t,ndim=2] A = np.empty( (window_size**2, len(exps)), dtype=np.float64 )
       for i, exp in enumerate( exps ):
          A[:,i] = (dx**exp[0]) * (dy**exp[1])

       # pad input array with appropriate values at the four borders
       cdef np.ndarray[np.float64_t,ndim=2] Z = np.zeros( (z.shape[0] + 2*half_size, z.shape[1] + 2*half_size), dtype=np.float64 )
       # top band
       Z[:half_size, half_size:-half_size] =  z[0, :] -  np.abs( np.flipud( z[1:half_size+1, :] ) - z[0, :] )
       # bottom band
       Z[-half_size:, half_size:-half_size] = z[-1, :]  + np.abs( np.flipud( z[-half_size-1:-1, :] )  -z[-1, :] )
       # left band
       Z[half_size:-half_size, :half_size] = np.tile( z[:,0].reshape(-1,1), [1,half_size]) - np.abs( np.fliplr( z[:, 1:half_size+1] ) - np.tile( z[:,0].reshape(-1,1), [1,half_size]) )
       # right band
       Z[half_size:-half_size, -half_size:] =  np.tile( z[:,-1].reshape(-1,1), [1,half_size] ) + np.abs( np.fliplr( z[:, -half_size-1:-1] ) - np.tile( z[:,-1].reshape(-1,1), [1,half_size] ) )
       # central band
       Z[half_size:-half_size, half_size:-half_size] = z

       # top left corner
       Z[:half_size,:half_size] = z[0,0] - np.abs( np.flipud(np.fliplr(z[1:half_size+1,1:half_size+1]) ) - z[0,0] )
       # bottom right corner
       Z[-half_size:,-half_size:] = z[-1,-1] + np.abs( np.flipud(np.fliplr(z[-half_size-1:-1,-half_size-1:-1]) ) - z[-1,-1] )

       # top right corner
       Z[:half_size,-half_size:] = Z[half_size,-half_size:] - np.abs( np.flipud(Z[half_size+1:2*half_size+1,-half_size:]) - Z[half_size,-half_size:] )
       # bottom left corner
       Z[-half_size:,:half_size] = Z[-half_size:,half_size].reshape(-1,1) - np.abs( np.fliplr(Z[-half_size:, half_size+1:2*half_size+1]) - Z[-half_size:,half_size].reshape(-1,1) )

       cdef np.ndarray[np.float64_t,ndim=2] m = np.zeros((window_size, window_size), dtype=np.float64)
       # solve system and convolve
       m = np.linalg.pinv(A)[0].reshape((window_size, -1))
       cdef np.ndarray[np.float64_t,ndim=2] out = np.zeros( np.shape(z), dtype=np.float64 )
       out = sp.fftconvolve(Z.astype('f'), m.astype('f'), mode='valid')
       self.data = out
       return
       
    # =========================================================
    @cython.boundscheck(False)
    @cython.cdivision(True)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cpdef np.ndarray[np.float64_t,ndim=2] getdata(self):
       '''
       Create a Savitsky-Golay digital filter from a 2D signal
       based on code from http://www.scipy.org/Cookbook/SavitzkyGolay

       Syntax
       ----------
       nr_pts = pysesa.sgolay.getdata()

       Parameters
       ----------
       self : instance
   	    pysesa.sgolay instance

       Returns
       -------
       self.data : ndarray, shape (N)
          the smoothed signal.
       '''
       return self.data  
